from asyncio.selector_events import BaseSelectorEventLoop

from .selector import EpollSelector

__all__ = ("WepollEventLoop",)

class WepollEventLoop(BaseSelectorEventLoop):
    """
    Selector event loop for wepoll.

    See `events.EventLoop <https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.EventLoop>`__ for API specification.
    """

    def __init__(self):
        super().__init__(EpollSelector())



from cpython.object cimport PyObject, PyTypeObject
from cpython.time cimport PyTime_t
from libc.stdint cimport uintptr_t

# After carefully observing socketmodule.h over from CPython and some 
# different versions of python I've determined it might be possible to 
# expose this data to the cython api by simply copying off socketmodule.h
# otherwise a simple array hack utilized form Cython shoud do the trick.

# Doing this allows use to extract the fd data sooner than later.

# Author: Vizonex

cdef extern from *:
    """
/* from socketmodule.h */
#define PySocket_MODULE_NAME    "_socket"
#define PySocket_CAPI_NAME      "CAPI"
#define PySocket_CAPSULE_NAME   PySocket_MODULE_NAME "." PySocket_CAPI_NAME

#ifdef MS_WINDOWS
/* typedef SOCKET SOCKET_T; */
// Only 1 change required so that we can ignore winsock.
typedef uintptr_t SOCKET_T;
#       ifdef MS_WIN64
#               define SIZEOF_SOCKET_T 8
#       else
#               define SIZEOF_SOCKET_T 4
#       endif
#else
typedef int SOCKET_T;
#       define SIZEOF_SOCKET_T SIZEOF_INT
#endif

/* Didn't need to recast the entire object only the parts that I thought were useful... */
typedef struct {
    PyObject_HEAD
    SOCKET_T sock_fd;
    int sock_family;
    int sock_type;
    int sock_proto;
    PyObject* (*errorhandler)();
    _PyTime_t sock_timeout;
} PySocketSockObject;

typedef struct  {
    PyTypeObject *Sock_Type;
    PyObject *error;
    PyObject *timeout_error;
} PySocketModule_APIObject;

/* ==== CYTHON PORT ==== */

PySocketModule_APIObject* __cython_socket_api;

/* Attempts to make pywepoll threadsafe */
PySocketModule_APIObject* cimport_socket(){
    return (PySocketModule_APIObject*)PyCapsule_Import(PySocket_CAPSULE_NAME, 0);
}

int import_socket(){
    /* PySocketModule_ImportModuleAndAPI Macro was not required 
     * we need the import to block incase of failure */ 
    __cython_socket_api = PyCapsule_Import(PySocket_CAPSULE_NAME, 0);
    return __cython_socket_api != NULL;
}


static inline int SocketType_CheckExact(PyObject* obj) {
    if (__cython_socket_api != NULL){
        return Py_IS_TYPE(obj, __cython_socket_api->Sock_Type);    
    }
    PyErr_SetString(PyExc_ImportError, "Required CPython Capsule _socket.CAPI was not imported");
    return -1;
}

static inline int SocketType_Check(PyObject* obj) {
    return SocketType_CheckExact(obj) || PyObject_TypeCheck(obj,  __cython_socket_api->Sock_Type);
}

static inline int Socket_GetFileDescriptor(PyObject* sock, uintptr_t *fd){
    if (!SocketType_Check(sock)){
        PyErr_SetString(PyExc_TypeError, "socket type is required");
        return -1;
    }
    *fd = ((PySocketSockObject*)sock)->sock_fd;
    return 0;
}
    """
    ctypedef uintptr_t SOCKET_T

    ctypedef struct PySocketSockObject:
        SOCKET_T sock_fd # Socket file descriptor
        int sock_family       # Address family, e.g., AF_INET
        int sock_type         # Socket type, e.g., SOCK_STREAM
        int sock_proto        # Protocol type, usually 0
        PyObject* (*errorhandler)() except NULL # Error handler; checks
                                         #   errno, returns NULL and
                                         #   sets a Python exception
        PyTime_t sock_timeout       # Operation timeout in seconds
                                        # 0.0 means non-blocking
    
    ctypedef struct PySocketModule_APIObject:
        PyTypeObject *Sock_Type
        PyObject *error
        PyObject *timeout_error
    
    int Socket_GetFileDescriptor(object sock, uintptr_t *fd)

    # Global
    PySocketModule_APIObject* __cython_socket_api 

    ctypedef class _socket.socket [object PySocketSockObject, check_size ignore]:
        cdef:
            # Reason for making the majortiy of these constant is to prevent fiddling
            # with sockets from python's end.

            @property
            cdef inline const int family(self):
                return self.sock_family
            
            @property
            cdef inline const int type(self):
                return self.sock_type
            
            @property
            cdef inline PyTime_t timeout(self):
                return self.sock_timeout

    int import_socket() except 0
    # imports _socket.CAPI this function is important to import first thing after cimport is finished
    
    int SocketType_CheckExact(object obj) except -1
    # checks if python type matches _socket.socket exactly
    # - 0 if False
    # - 1 if True 
    # - -1 if CAPI did not import raises ImportError

    int SocketType_Check(object obj) except -1
    # checks if python type is socket.socket or a subclass 
    # - 0 if False
    # - 1 if True 
    # - -1 if object is NULL

    PySocketModule_APIObject* cimport_socket() except NULL



from wepoll import epoll, EPOLLIN, EPOLLOUT
import pytest

cyares = pytest.importorskip("cyares")
if cyares is not None:
    from cyares import Channel

# based off pycares's testsuite

READ = EPOLLIN
WRITE = EPOLLOUT

class TestCyaresWepoll:
    channel: "Channel"

    def wait(self):
        # The function were really testing is this wait function

        while self.channel.running_queries:
            timeout = self.channel.timeout()
            if timeout == 0.0:
                self.channel.process_no_fds()
                continue
            for fd, event in self.poll.poll(timeout):
                if event & ~EPOLLIN:
                    self.channel.process_write_fd(fd)
                if event & ~EPOLLOUT:
                    self.channel.process_read_fd(fd)

    def socket_state_cb(self, fd: int, r:bool, w: bool):
        flags = 0
        if r:
            flags |= READ
        if w:
            flags |= WRITE
        
        if flags:
            self.poll.register(fd, flags)
        else:
            self.poll.unregister(fd)

    def test_resolve(self):
        self.poll = epoll()
        self.channel = Channel(event_thread=False, servers=["8.8.8.8", "8.8.4.4"], sock_state_cb=self.socket_state_cb)
        fut = self.channel.query("python.org", "A")
        self.wait()
        assert fut.result()



from wepoll import WepollEventLoop
from cyares.aio import DNSResolver
import pytest

@pytest.fixture(
    params=[
        pytest.param(
            ("asyncio", {"loop_factory": WepollEventLoop}),
            id="asyncio[WepollEventLoop]",
        )
    ]
)
def anyio_backend(request: pytest.FixtureRequest):
    return request.param


@pytest.mark.anyio
async def test_dns_resolver_over_wepoll() -> None:
    async with DNSResolver(["8.8.8.8", "8.8.4.4"], event_thread=False) as dns:
        assert await dns.query("google.com", "A")



# cython: freethreading_compatible = True
cimport cython
from cpython.exc cimport (
    PyErr_CheckSignals, 
    PyErr_SetFromErrno,
    PyErr_SetFromWindowsErr, 
    PyErr_SetObject
)
from cpython.long cimport PyLong_Check
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.object cimport PyObject_TypeCheck
from cpython.time cimport PyTime_t, monotonic_ns
from cpython.type cimport PyType_GenericAlloc
from libc.limits cimport INT_MAX, INT_MIN

from .msvcrt cimport get_osfhandle
from .socket cimport cimport_socket, socket
from .wepoll cimport *

# Inspired by the original 2006 cython epoll twisted code and CPython's vesion


# TODO: make C-API Capsule in a later update and then migrate to CPython 
# so that the cython functionality can at least be retained.





cdef extern from "Python.h":
    """

/* Dear CPython devs, please stop having me try to force my hand 
 * by performing Code injections I don't like this at all. */

// 3.13 is cursed...
#if (PY_VERSION_HEX >= 0x030d00f0)

/* I hate this */
#include <math.h>

typedef int64_t _PyTime_t;

#ifndef PyTime_MIN
    #define PyTime_MIN INT64_MIN
#endif
#ifndef PyTime_MAX
    #define PyTime_MAX INT64_MAX
#endif

#define SEC_TO_MS 1000

/* To microseconds (10^-6) */
#define MS_TO_US 1000
#define SEC_TO_US (SEC_TO_MS * MS_TO_US)

/* To nanoseconds (10^-9) */
#define US_TO_NS 1000
#define MS_TO_NS (MS_TO_US * US_TO_NS)
#define SEC_TO_NS (SEC_TO_MS * MS_TO_NS)

/* Conversion from nanoseconds */
#define NS_TO_MS (1000 * 1000)
#define NS_TO_US (1000)
#define NS_TO_100NS (100)

typedef enum {
    // Round towards minus infinity (-inf).
    // For example, used to read a clock.
    _PyTime_ROUND_FLOOR=0,

    // Round towards infinity (+inf).
    // For example, used for timeout to wait "at least" N seconds.
    _PyTime_ROUND_CEILING=1,

    // Round to nearest with ties going to nearest even integer.
    // For example, used to round from a Python float.
    _PyTime_ROUND_HALF_EVEN=2,

    // Round away from zero
    // For example, used for timeout. _PyTime_ROUND_CEILING rounds
    // -1e-9 to 0 milliseconds which causes bpo-31786 issue.
    // _PyTime_ROUND_UP rounds -1e-9 to -1 millisecond which keeps
    // the timeout sign as expected. select.poll(timeout) must block
    // for negative values.
    _PyTime_ROUND_UP=3,

    // _PyTime_ROUND_TIMEOUT (an alias for _PyTime_ROUND_UP) should be
    // used for timeouts.
    _PyTime_ROUND_TIMEOUT = _PyTime_ROUND_UP
} _PyTime_round_t;


static void
pytime_time_t_overflow(void)
{
    PyErr_SetString(PyExc_OverflowError,
                    "timestamp out of range for platform time_t");
}


static void
pytime_overflow(void)
{
    PyErr_SetString(PyExc_OverflowError,
                    "timestamp too large to convert to C PyTime_t");
}

static double
pytime_round_half_even(double x)
{
    double rounded = round(x);
    if (fabs(x-rounded) == 0.5) {
        /* halfway case: round to even */
        rounded = 2.0 * round(x / 2.0);
    }
    return rounded;
}

static double
pytime_round(double x, _PyTime_round_t round)
{
    /* volatile avoids optimization changing how numbers are rounded */
    volatile double d;

    d = x;
    if (round == _PyTime_ROUND_HALF_EVEN) {
        d = pytime_round_half_even(d);
    }
    else if (round == _PyTime_ROUND_CEILING) {
        d = ceil(d);
    }
    else if (round == _PyTime_ROUND_FLOOR) {
        d = floor(d);
    }
    else {
        assert(round == _PyTime_ROUND_UP);
        d = (d >= 0.0) ? ceil(d) : floor(d);
    }
    return d;
}


static int
pytime_from_double(_PyTime_t *tp, double value, _PyTime_round_t round,
                   long unit_to_ns)
{
    /* volatile avoids optimization changing how numbers are rounded */
    volatile double d;

    /* convert to a number of nanoseconds */
    d = value;
    d *= (double)unit_to_ns;
    d = pytime_round(d, round);

    /* See comments in pytime_double_to_denominator */
    if (!((double)PyTime_MIN <= d && d < -(double)PyTime_MIN)) {
        pytime_time_t_overflow();
        *tp = 0;
        return -1;
    }
    _PyTime_t ns = (_PyTime_t)d;

    *tp = ns;
    return 0;
}

static inline int
pytime_mul_check_overflow(_PyTime_t a, _PyTime_t b)
{
    if (b != 0) {
        assert(b > 0);
        return ((a < PyTime_MIN / b) || (PyTime_MAX / b < a));
    }
    else {
        return 0;
    }
}


static inline int
pytime_mul(_PyTime_t *t, _PyTime_t k)
{
    assert(k >= 0);
    if (pytime_mul_check_overflow(*t, k)) {
        *t = (*t >= 0) ? PyTime_MAX : PyTime_MIN;
        return -1;
    }
    else {
        *t *= k;
        return 0;
    }
}


static int
pytime_from_object(_PyTime_t *tp, PyObject *obj, _PyTime_round_t round,
                   long unit_to_ns)
{
    if (PyFloat_Check(obj)) {
        double d;
        d = PyFloat_AsDouble(obj);
        if (isnan(d)) {
            PyErr_SetString(PyExc_ValueError, "Invalid value NaN (not a number)");
            return -1;
        }
        return pytime_from_double(tp, d, round, unit_to_ns);
    }

    long long sec = PyLong_AsLongLong(obj);
    if (sec == -1 && PyErr_Occurred()) {
        if (PyErr_ExceptionMatches(PyExc_OverflowError)) {
            pytime_overflow();
        }
        else if (PyErr_ExceptionMatches(PyExc_TypeError)) {
            PyErr_Format(PyExc_TypeError,
                         "'%T' object cannot be interpreted as an integer or float",
                         obj);
        }
        return -1;
    }

    static_assert(sizeof(long long) <= sizeof(_PyTime_t),
                  "PyTime_t is smaller than long long");
    _PyTime_t ns = (_PyTime_t)sec;
    if (pytime_mul(&ns, unit_to_ns) < 0) {
        pytime_overflow();
        return -1;
    }

    *tp = ns;
    return 0;
}


int
_PyTime_FromSecondsObject(_PyTime_t *tp, PyObject *obj, _PyTime_round_t round)
{
    return pytime_from_object(tp, obj, round, SEC_TO_NS);
}
 
#endif

typedef _PyTime_round_t PyTime_round_t;
#define PyTime_FromSecondsObject _PyTime_FromSecondsObject

/* From Python 3.14 Brought here for backwards comptability*/

#ifndef PyTime_MAX
    #define PyTime_MAX LLONG_MAX
#endif
#ifndef PyTime_MIN
    #define PyTime_MIN LLONG_MIN
#endif

static inline int
compat_pytime_add(int64_t *t1, int64_t t2)
{
    if (t2 > 0 && *t1 > PyTime_MAX - t2) {
        *t1 = PyTime_MAX;
        return -1;
    }
    else if (t2 < 0 && *t1 < PyTime_MIN - t2) {
        *t1 = PyTime_MIN;
        return -1;
    }
    else {
        *t1 += t2;
        return 0;
    }
}

int64_t
_PyTime_Add(int64_t t1, int64_t t2)
{
    (void)compat_pytime_add(&t1, t2);
    return t1;
}
#define PyTime_Add _PyTime_Add
"""
   
    enum _PyTime_round_t:
        # Round towards minus infinity (-inf).
        # For example, used to read a clock.
        _PyTime_ROUND_FLOOR = 0,
        # Round towards infinity (+inf).
        # For example, used for timeout to wait "at least" N seconds.
        _PyTime_ROUND_CEILING = 1,
        # Round to nearest with ties going to nearest even integer.
        # For example, used to round from a Python float.
        _PyTime_ROUND_HALF_EVEN =2 ,
        # Round away from zero
        # For example, used for timeout. _PyTime_ROUND_CEILING rounds
        # -1e-9 to 0 milliseconds which causes bpo-31786 issue.
        _PyTime_ROUND_UP = 3
        #    the timeout sign as expected. select.poll(timeout) must block
        #    for negative values."
        # _PyTime_ROUND_TIMEOUT (an alias for _PyTime_ROUND_UP) should be
        # used for timeouts.
        _PyTime_ROUND_TIMEOUT = 3

    ctypedef _PyTime_round_t PyTime_round_t
    # Comptable version of _PyTime_Add for older versions of Python
    PyTime_t PyTime_Add(PyTime_t t1, PyTime_t t2)
    int PyTime_FromSecondsObject(PyTime_t *t, object obj, PyTime_round_t round)
    PyTime_t _PyTime_AsMilliseconds(PyTime_t timeout, PyTime_round_t round)

# Recasts so that my Cyright extension doesn't misbehave
cdef PyTime_round_t PyTime_ROUND_TIMEOUT = _PyTime_ROUND_TIMEOUT
cdef PyTime_round_t PyTime_ROUND_FLOOR = _PyTime_ROUND_FLOOR
cdef PyTime_round_t PyTime_ROUND_UP = _PyTime_ROUND_UP
cdef PyTime_round_t PyTime_ROUND_CEILING = _PyTime_ROUND_CEILING
cdef PyTime_round_t PyTime_ROUND_HALF_EVEN = _PyTime_ROUND_HALF_EVEN



cdef extern from "windows.h" nogil:
    pass

cdef extern from "handleapi.h" nogil:
    """
#ifndef HANDLE_FLAG_INHERIT
#define HANDLE_FLAG_INHERIT 0x00000001
#endif
    """
    ctypedef unsigned long DWORD
    bint SetHandleInformation(
        HANDLE hObject,
        DWORD  dwMask,
        DWORD  dwFlags
    )
    DWORD HANDLE_FLAG_INHERIT

cdef extern from "errno.h" nogil:
    """
#define wepoll_set_errno(err) errno = err
    """
    cdef int errno
    cdef char *strerror(int)
    cdef int EINTR
    void wepoll_set_errno(int)

DEF FD_SETSIZE = 512


 
# Keep final the same way select does on linux
@cython.final
cdef class epoll:
    # internal methods first then try mimicing python
    # doing so this way allows us to create a 
    # cpython capsule if we wish...

    cdef SOCKET _fd_from_object(self, object obj) except -1:
        # Made threasafe in 0.1.3 by only making init need to cimport the capsule.
        if PyObject_TypeCheck(obj, self.socket_api.Sock_Type):
            return (<socket>obj).sock_fd
        elif PyLong_Check(obj):
            return <SOCKET>obj
        
        PyErr_SetObject(TypeError, f"{obj!r} not supported")
        return -1
    
    # would've used nogil but it did not feel as clean as PyEval was
    cdef int _create(self, int sizehint) except -1:
        with nogil:
            self.handle = epoll_create(sizehint)
        return -1 if self.handle == NULL else 0

    # TODO: Deprecate create1 in future update.
    cdef int _create1(self) except -1: 
        with nogil:
            self.handle = epoll_create1(0)
        return -1 if self.handle == NULL else 0
 
    cdef int _close(self):
        cdef int ret
        with nogil:
            ret = epoll_close(self.handle)
            if ret < 0:
                wepoll_set_errno(ret)
        return ret
    
    # TODO: In the future I will provide a way to make it so that 
    # other types of data besides just sockets can get polled.

    cdef int _ctl(self, int op, SOCKET sock, epoll_event* event):
        cdef int ret
        with nogil:
            ret = epoll_ctl(self.handle, op, sock, event)

        if ret < 0:
            PyErr_SetFromErrno(OSError)
            return -1
        return ret

    cdef int _wait(self, epoll_event* events, int maxevents, int timeout):
        with nogil:
            return epoll_wait(self.handle, events, maxevents, timeout)
    
    cdef int _init(self, int sizehint, HANDLE handle):
        if handle == NULL:
            if sizehint > 0:
                self._create(sizehint)
            else:
                self._create1()
                # optimzed version of _Py_set_inheritable for windows
                if not SetHandleInformation(self.handle, HANDLE_FLAG_INHERIT, 0):
                    PyErr_SetFromWindowsErr(0)
                    return -1
        else:
            self.closed = 0
            self.handle = handle
        return 0


    cdef int _handle_ctl_result(self, int result) except -1:
        if result < 0:
            wepoll_set_errno(result)
            PyErr_SetFromErrno(OSError)
            return -1
        return 0
    
    cdef int _pools_closed(self) except -1:
        if self.handle == NULL or self.closed:
            # Pools closed due to aids
            PyErr_SetObject(RuntimeError, "I/O operation on closed epoll object")
            return -1
        return 0
 
    # NOTE Flags are deprecated in select standard library so no point in using it here...
    def __init__(self, int sizehint = -1):
        if sizehint == -1:
            sizehint = FD_SETSIZE - 1
        
        elif sizehint <= 0:
            raise ValueError("negative sizehint")
        
        # NOTE: Throw me an issue if a memory leak is noticable.
        # I will then be able to diagnose if this was the reason.
        self.socket_api = cimport_socket()
        if self.socket_api == NULL:
            raise

        if self._init(sizehint, NULL) < 0:
            raise

    cpdef uintptr_t fileno(self):
        """Return the file descriptor number of the control fd.
        on windows this recasts the handle to a `uintptr_t` to a 
        PyLongObject (int)."""
        return <uintptr_t>self.handle


    cpdef object close(self):
        """Close the control file descriptor of the epoll object.

        Raises:
            OSError: if close fails.
        """
        cdef int _errno
        if not self.closed:
            _errno = self._close()
            if _errno < 0:
                wepoll_set_errno(_errno)
                PyErr_SetFromErrno(OSError)
                raise
            self.closed = True
    
    cpdef object register(self, object fd, unsigned int eventmask):
        cdef epoll_event ev
        cdef SOCKET _fd

        if self._pools_closed() < 0:
            raise

        _fd = ev.data.sock = self._fd_from_object(fd)
        ev.events = eventmask
        if self._ctl(EPOLL_CTL_ADD, _fd, &ev) < 0:
            raise
            
    cpdef object modify(self, object fd, unsigned int eventmask):
        """Modify a registered file descriptor.
        
        Parameters
        ----------

        :param fd: a `socket` or `int` object of a file descriptor
        :param eventmask: A List of epoll flags to set for the modified fd.

        Raises:
            TypeError: if obtaining the file descriptor from the 
                python object fails or is an invalid type.

            RuntimeError: if epoll is closed.

        """
        cdef epoll_event ev
        cdef SOCKET _fd

        if self._pools_closed() < 0:
            raise
        _fd = ev.data.sock = self._fd_from_object(fd)
        ev.events = eventmask
        if self._ctl(EPOLL_CTL_MOD, _fd, &ev) < 0:
            raise

    cpdef object unregister(self, object fd):
        """
        Remove a registered file descriptor from the epoll object.

        Parameters
        ----------

        :param fd: a `socket` or `int` object of a file descriptor


        Raises:
            OSError: if unregistering the file descriptor fails.

        """
        cdef epoll_event ev
        if self._pools_closed() < 0:
            raise

        if self._ctl(EPOLL_CTL_DEL, <SOCKET>self._fd_from_object(fd), &ev) < 0:
            raise

    cpdef list poll(self, object timeout = None, int maxevents = -1):
        """
        Wait for events. timeout in seconds (float)
        
        Parameters
        ----------

        :param timeout: timeout in seconds (float or int)
        :param maxevent: maximum number of events to listen for default is -1 which will be a varaious amount.
            Passing -1 to maxevents is discouraged and should be left alone instead if chosen to omit or ignore
        Raises:
            TypeError: if timeout type is not supported
            ValueError: if maxevents is less than -1 


        """
        cdef PyTime_t _timeout, deadline, ms
        cdef epoll_event *evs = NULL
        cdef int nfds, i
        cdef void* handle = self.handle

        _timeout = deadline = -1
        nfds = 0
        if timeout is not None:
            if PyTime_FromSecondsObject(&_timeout, timeout, PyTime_ROUND_TIMEOUT) < 0:
                raise TypeError(f"{timeout!r} not supported")

            # add timeout by current time as a monotonic clock using Python's
            # Normal API So that we can mimic linux's python epoll system. 
            
            ms = _PyTime_AsMilliseconds(_timeout, _PyTime_ROUND_CEILING)
            if ms < INT_MIN or ms > INT_MAX:
                raise OverflowError("timeout is too large")
            if ms < 0:
                ms = -1
            
            if _timeout >= 0:
                deadline = PyTime_Add(monotonic_ns(), _timeout)
            

        if maxevents == -1:
            maxevents = FD_SETSIZE - 1
        elif maxevents < 1:
            raise ValueError(f"maxevents must be greater than 0, got {maxevents}")

        evs = <epoll_event*>PyMem_Malloc(sizeof(epoll_event) * maxevents)
        if evs == NULL:
            raise MemoryError
        try:
            while True:
                with nogil:
                    errno = 0
                    nfds = epoll_wait(handle, evs, maxevents, <int>ms)

                if nfds > 0:
                    break

                elif nfds < 0:
                    PyErr_SetFromErrno(OSError)
                    raise

                # check for ctrl+C from end user wanting to stop program
                if PyErr_CheckSignals() < 0:
                    raise

                if timeout >= 0:
                    timeout = deadline - monotonic_ns()
                    if (timeout < 0):
                        nfds = 0
                        break
                    ms = _PyTime_AsMilliseconds(timeout, _PyTime_ROUND_CEILING)

            return [(evs[i].data.fd, evs[i].events) for i in range(nfds)]
        finally:
            PyMem_Free(evs)
        
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
    
    def __dealloc__(self):
        self.close()
    
    @classmethod
    def fromfd(cls, int fd):
        """creates a epoll (wepoll) from msvcrt using _get_osfhandle(fd) 
        from `io.h` in C to get the file descriptor's handle"""
        cdef epoll poll = PyType_GenericAlloc(cls, 0)
        cdef void* handle

        if get_osfhandle(&handle, fd) < 0:
            # Reraise the OS Error we had setup previously...
            raise 

        if poll._init(FD_SETSIZE - 1, handle) < 0:
            raise
        return poll    



from ._wepoll import epoll
from .flags import EPOLLERR as EPOLLERR
from .flags import EPOLLHUP as EPOLLHUP
from .flags import EPOLLIN as EPOLLIN
from .flags import EPOLLMSG as EPOLLMSG
from .flags import EPOLLONESHOT as EPOLLONESHOT
from .flags import EPOLLOUT as EPOLLOUT
from .flags import EPOLLPRI as EPOLLPRI
from .flags import EPOLLRDBAND as EPOLLRDBAND
from .flags import EPOLLRDHUP as EPOLLRDHUP
from .flags import EPOLLRDNORM as EPOLLRDNORM
from .flags import EPOLLWRBAND as EPOLLWRBAND
from .flags import EPOLLWRNORM as EPOLLWRNORM
from .loop import WepollEventLoop
from .selector import EpollSelector

__author__ = "Vizonex"
__version__ = "0.3.3"
__all__ = (
    "EPOLLERR",
    "EPOLLHUP",
    "EPOLLIN",
    "EPOLLMSG",
    "EPOLLONESHOT",
    "EPOLLOUT",
    "EPOLLPRI",
    "EPOLLRDHUP",
    "EPOLLWRBAND",
    "EPOLLWRNORM",
    "EpollSelector",
    "WepollEventLoop",
    "__author__",
    "__version__",
    "epoll",
)






from .socket cimport PySocketModule_APIObject
from .wepoll cimport *


# NOTE: Many portions of code can be externally 
# used in cython hence it's creation & setup
cdef class epoll:
    cdef:
        HANDLE handle
        readonly bint closed
        # Attempts to make socketmodule threadsafe during use...
        PySocketModule_APIObject* socket_api

    cdef SOCKET _fd_from_object(self, object obj) except -1
    cdef int _create(self, int sizehint) except -1
    cdef int _create1(self) except -1
    cdef int _close(self)
    cdef int _ctl(self, int op, SOCKET sock, epoll_event* event)
    cdef int _wait(self, epoll_event* events, int maxevents, int timeout)
    cdef int _init(self, int sizehint, HANDLE handle)
    cdef int _handle_ctl_result(self, int result) except -1
    cdef int _pools_closed(self) except -1
    cpdef object close(self)
    cpdef object register(self, object fd, unsigned int eventmask)
    cpdef object modify(self, object fd, unsigned int eventmask)
    cpdef object unregister(self, object fd)
    cpdef list poll(self, object timeout =*, int maxevents =*)
    
    cpdef uintptr_t fileno(self)
    



# === FLAGS ===

# SEE: https://docs.python.org/3/library/select.html#edge-and-level-trigger-polling-epoll-objects
# NOTE: Some flags might not be supported such as EPOLLET, EPOLLEXCLUSIVE

EPOLLIN = 1 << 0
"""Available for read"""
EPOLLPRI = 1 << 1
"""Urgent data for read"""
EPOLLOUT = 1 << 2
"""Available for write"""
EPOLLERR = 1 << 3
"""Error condition happened on the assoc. fd"""
EPOLLHUP = 1 << 4
"""Hang up happened on the assoc. fd"""
EPOLLRDNORM = 1 << 6
"""Equivalent to EPOLLIN"""
EPOLLRDBAND = 1 << 7
"""Priority data band can be read."""
EPOLLWRNORM = 1 << 8
"""Equivalent to EPOLLOUT"""
EPOLLWRBAND = 1 << 9
"""Priority data may be written."""
EPOLLMSG = 1 << 10  # Never Reported
"""Ignored."""
EPOLLRDHUP = 1 << 13
"""Stream socket peer closed connection or shut down writing half of connection."""
EPOLLONESHOT = 1 << 31
"""Set one-shot behavior. After one event is pulled out, the fd is internally disabled"""

__all__ = (
    "EPOLLERR",
    "EPOLLHUP",
    "EPOLLIN",
    "EPOLLMSG",
    "EPOLLONESHOT",
    "EPOLLOUT",
    "EPOLLPRI",
    "EPOLLRDHUP",
    "EPOLLWRBAND",
    "EPOLLWRNORM",
)



import sys
from socket import socket
from types import TracebackType
from typing import final

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

@final
class epoll:
    def __init__(self, sizehint: int = ...) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = ...,
        exc_tb: TracebackType | None = None,
        /,
    ) -> None: ...
    def close(self) -> None: ...
    closed: bool
    def fileno(self) -> int: ...
    def register(self, fd: int | socket, eventmask: int = ...) -> None: ...
    def modify(self, fd: int | socket, eventmask: int) -> None: ...
    def unregister(self, fd: int | socket) -> None: ...
    def poll(
        self, timeout: float | None = None, maxevents: int = -1
    ) -> list[tuple[int, int]]:
        """
        Wait for events. timeout in seconds (float)

        Parameters
        ----------

        :param timeout: timeout in seconds (float or int)
        :param maxevent: maximum number of events to listen for default is -1 which will be a varaious amount.
            Passing -1 to maxevents is discouraged and should be left alone instead if chosen to omit or ignore
        Raises:
            TypeError: if timeout type is not supported
            ValueError: if maxevents is less than -1
        """

    # Maybe in a Future release but we shall see...
    @classmethod
    def fromfd(cls, fd: int, /) -> epoll:
        """creates a epoll (wepoll) from msvcrt using _get_osfhandle(fd)
        from `io.h` in C to get the file descriptor's handle"""



import sys
from selectors import SelectorKey
from selectors import _PollLikeSelector as PollLikeSelector
from typing import TYPE_CHECKING

from ._wepoll import epoll
from .flags import EPOLLIN, EPOLLOUT

__all__ = ("EpollSelector",)

# Added here if you didn't want to grab from somewhere else
EVENT_READ = 1  # (1 << 0)
EVENT_WRITE = 2  # (1 << 1)


# 3.13 compatabaility flags
_NOT_EPOLLIN = ~EPOLLIN
_NOT_EPOLLOUT = ~EPOLLOUT

class EpollSelector(PollLikeSelector):
    """Wepoll-based selector for windows operating systems"""

    _selector_cls = epoll
    _EVENT_READ = EPOLLIN
    _EVENT_WRITE = EPOLLOUT

    # I added this in as an obvious optimization
    __slots__ = ("_selector", "_fd_to_key", "_map")

    if TYPE_CHECKING:
        _selector: epoll
        if sys.version_info < (3, 13):
            _fd_to_key: "dict[int, SelectorKey]"

    if sys.version_info < (3, 13):
        # XXX: I don't know if this will work on later versions of
        # Python I haven't tried it yet...
        def select(self, timeout: float | None = None) -> list[tuple[int, SelectorKey]]:
            # This is shared between poll() and epoll().
            # epoll() has a different signature and handling of timeout parameter.
            if timeout and (timeout <= 0):
                timeout = 0
            
            ready: list[tuple[SelectorKey, int]] = []
            try:
                fd_event_list = self._selector.poll(timeout)
            except InterruptedError:
                return ready

            fd_to_key = self._fd_to_key
            for fd, event in fd_event_list:
                # We can do better than what is in python's standard selectors library...
                if key := fd_to_key.get(fd):
                    events = 0
                    if event & ~self._EVENT_READ:
                        events |= EVENT_WRITE
                    if event & ~self._EVENT_WRITE:
                        events |= EVENT_READ
                    ready.append((key, events & key.events))
            return ready
    else:
        def select(self, timeout: float | None = None) -> list[tuple[int, SelectorKey]]:
            if timeout is None:
                timeout = -1
            elif timeout <= 0:
                timeout = 0

            # epoll_wait() expects `maxevents` to be greater than zero;
            # we want to make sure that `select()` can be called when no
            # FD is registered.
            max_ev = len(self._fd_to_key) or 1

            ready: list[tuple[SelectorKey, int]] = []
            try:
                fd_event_list = self._selector.poll(timeout, max_ev)
            except InterruptedError:
                return ready

            fd_to_key = self._fd_to_key
            for fd, event in fd_event_list:
                if key := fd_to_key.get(fd):
                    events = (event & _NOT_EPOLLIN and EVENT_WRITE) | (
                        event & _NOT_EPOLLOUT and EVENT_READ
                    )
                    ready.append((key, events & key.events))
            return ready



from Cython.Build import cythonize
from setuptools import Extension, setup

# TODO: Move to pyproject.toml setup like in pyduktape3 somewhere around 0.3.0
setup(
    ext_modules=cythonize(
        [
            Extension(
                "wepoll._wepoll",
                ["wepoll/_wepoll.pyx", "vendor/wepoll/wepoll.c"],
                include_dirs=["vendor/wepoll"],
                libraries=["advapi32", "iphlpapi", "psapi", "ws2_32"],
            )
        ]
    )
)



import selectors
import tempfile
import unittest
from test.test_selectors import ScalableSelectorMixIn, BaseSelectorTestCase
from wepoll import EpollSelector

# Code is borrowed from python's testsuite to ensure wepoll matches up with unix epolls


class EpollSelectorTestCase(
    BaseSelectorTestCase, ScalableSelectorMixIn, unittest.TestCase
):
    SELECTOR = EpollSelector

    def test_modify_unregister(self):
        if self.SELECTOR.__name__ == "EpollSelector":
            patch = unittest.mock.patch("wepoll.EpollSelector._selector_cls")
        else:
            raise self.skipTest("")

        with patch as m:
            m.return_value.modify = unittest.mock.Mock(side_effect=ZeroDivisionError)
            s = self.SELECTOR()
            self.addCleanup(s.close)
            rd, wr = self.make_socketpair()
            s.register(rd, selectors.EVENT_READ)
            self.assertEqual(len(s._map), 1)
            with self.assertRaises(ZeroDivisionError):
                s.modify(rd, selectors.EVENT_WRITE)
            self.assertEqual(len(s._map), 0)

    def test_register_file(self):
        # epoll(7) returns EPERM when given a file to watch
        s = self.SELECTOR()
        with tempfile.NamedTemporaryFile() as f:
            with self.assertRaises(IOError):
                s.register(f, selectors.EVENT_READ)
            # the SelectorKey has been removed
            with self.assertRaises(KeyError):
                s.get_key(f)
    
    def test_empty_select(self):
        # Issue #23009: Make sure EpollSelector.select() works when no FD is
        # registered.
        s = self.SELECTOR()
        self.addCleanup(s.close)
        self.assertEqual(s.select(timeout=0), [])
